from typing import *
import pathlib

import pytest
from pytest import mark
from pytest_aux import *

from funcs_aux import *


# =====================================================================================================================
class Test__FromVariants:
    @classmethod
    def setup_class(cls):
        pass
        cls.Victim = type("Victim", (ValueVariants,), {})

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__val_def__reset(self):
        victim = self.Victim(value="var1", variants=["VAR1", "VAR2"])
        assert victim.VALUE == "VAR1"
        assert victim.VALUE != "VAR2"
        assert victim.VALUE_DEFAULT == "var1"

        victim.VALUE = "var2"
        assert victim.VALUE != "VAR1"
        assert victim.VALUE == "VAR2"
        assert victim.VALUE_DEFAULT == "var1"

        victim.reset()
        assert victim.VALUE == "VAR1"
        assert victim.VALUE != "VAR2"
        assert victim.VALUE_DEFAULT == "var1"

    def test__double_objects(self):
        victim1 = self.Victim(value="var1", variants=["VAR1", "VAR11"])
        victim2 = self.Victim(value="var2", variants=["VAR2", "VAR22"])
        assert victim1.VALUE == "VAR1"
        assert victim2.VALUE == "VAR2"

        assert victim1.VALUE != "VAR11"
        assert victim2.VALUE != "VAR22"

        victim1.VALUE = "VAR11"
        victim2.VALUE = "VAR22"

        assert victim1.VALUE != "VAR1"
        assert victim2.VALUE != "VAR2"

        assert victim1.VALUE == "VAR11"
        assert victim2.VALUE == "VAR22"

        try:
            victim1.VALUE = "VAR2"
            assert False
        except:
            assert True

    def test__case(self):
        victim = self.Victim(value="var1", variants=["VAR1", "VAR2"])
        assert victim.VALUE == "VAR1"
        assert str(victim) == "VAR1"

        try:
            victim = self.Victim(value="var1", variants=["VAR1", "VAR2"], case_insensitive=False)
            assert False
        except:
            assert True

    def test__variants_validate(self):
        victim = self.Victim(value="var", variants=["VAR", "var"], case_insensitive=False)
        assert victim.VALUE == "var"

        try:
            victim = self.Victim(value="var123", variants=["VAR", "var"], case_insensitive=False)
            assert False
        except Exx__ValueNotInVariants:
            pass
        except Exception as exx:
            print(f"{exx!r}")
            assert False

        try:
            victim = self.Victim(value="var123", variants=["VAR", "var"], case_insensitive=True)
            assert False
        except Exx__VariantsIncompatible:
            pass
        except Exception as exx:
            print(f"{exx!r}")
            assert False

    def test__types__None(self):
        victim = self.Victim(variants=["NONE", ])
        assert victim.VALUE == Value_NotPassed
        # assert str(victim) == "NONE"

        victim = self.Victim(value=None, variants=["NONE", ])
        assert victim.VALUE == "NONE"
        assert str(victim) == "NONE"

        victim = self.Victim(value="None", variants=[None, ])
        assert victim.VALUE is None
        assert str(victim) == "None"

    def test__types__int(self):
        victim = self.Victim(value=1, variants=["1", ])
        assert victim.VALUE == "1"
        assert str(victim) == "1"

        victim = self.Victim(value="1", variants=[1, ])
        assert victim.VALUE == 1
        assert str(victim) == "1"

    def test__cmp__same_obj(self):
        assert self.Victim(value=None, variants=["NONE", ]) == self.Victim(value="None", variants=["NONE", ])
        assert self.Victim(value="NONE", variants=["NONE", ]) == self.Victim(value="None", variants=["NONE", ])
        assert self.Victim(value=None, variants=["None", ]) == self.Victim(value=None, variants=["NONE", ])

        assert self.Victim(value=None, variants=["None", ], case_insensitive=False) != self.Victim(value=None, variants=["NONE", ])

    def test__cmp__simple_value(self):
        assert self.Victim(value=None, variants=["NONE", ]) == "NONE"
        assert self.Victim(value="NONE", variants=["NONE", ]) == "NONE"
        assert self.Victim(value=None, variants=["None", ]) == "NONE"
        assert self.Victim(value=None, variants=[None, ]) == "NONE"
        assert self.Victim(value=None, variants=[None, ], case_insensitive=False) != "NONE"
        assert self.Victim(value=None, variants=[None, ], case_insensitive=False) == "None"

    def test__contain(self):
        victim = self.Victim(variants=["NONE", ])
        assert None in victim
        assert "None" in victim
        assert "NONE" in victim

    def test__len(self):
        assert len(self.Victim(variants=[0, ])) == 1
        assert len(self.Victim(variants=[0, 1])) == 2

    def test__iter(self):
        assert list(self.Victim(variants=[0, ])) == [0, ]
        assert list(self.Victim(variants=[0, 1])) == [0, 1, ]


# =====================================================================================================================
